import numpy as np
from traitlets import HasTraits, Instance
from vedo import Plotter
from vedo.pyplot import histogram
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from regexport.model import AppState
from regexport.views.utils import HasWidget


class HistogramModel(HasTraits):
    x_data = Instance(np.ndarray, default_value=np.zeros(0))
    y_data = Instance(np.ndarray, default_value=np.zeros(0))
    z_data = Instance(np.ndarray, default_value=np.zeros(0))
    x_hist_title = "X"


    def register(self, model: AppState):
        self.model = model
        model.observe(self.update, ['cells'])

    def update(self, change):
        if (cells := self.model.cells) is None:
            self.x_data = np.zeros(0)
            self.y_data = np.zeros(0)
            self.z_data = np.zeros(0)
        else:
            self.x_data = cells.X.values
            self.y_data = cells.Y.values
            self.z_data = cells.Z.values


class HistogramView(HasWidget):

    def __init__(self, model: HistogramModel):

        widget = QVTKRenderWindowInteractor()
        HasWidget.__init__(self, widget=widget)
        self.plotter = Plotter(qtWidget=widget)

        self.model = model
        self.model.observe(self.render)

    def render(self, change):
        hist = histogram(self.model.y_data, bins=20, title=self.model.x_hist_title)
        self.plotter.show(hist, mode=12)



from functools import partial
from pathlib import Path
from typing import Optional

import numpy as np
from PyQt5.QtCore import QThreadPool
from vedo import Plotter, Points, Mesh
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from regexport.utils.parallel import Task
from regexport.utils.profiling import warn_if_slow
from regexport.views.plot_3d import PlotterModel
from regexport.views.utils import HasWidget


class PlotterView(HasWidget):

    def __init__(self, model: PlotterModel):
        self.model = model
        self.item_points = {}
        self._atlas_mesh = None

        widget = QVTKRenderWindowInteractor()
        HasWidget.__init__(self, widget=widget)
        self.plotter = Plotter(qtWidget=widget)

        self.model.observe(self.observe_atlas_mesh, ['atlas_mesh'])
        self.model.observe(self.render, ['points'])

    @property
    def atlas_mesh(self) -> Optional[Mesh]:
        return self._atlas_mesh

    @atlas_mesh.setter
    def atlas_mesh(self, value: Optional[Mesh]):
        self._atlas_mesh = value
        self.render(None)

    @staticmethod
    def load_mesh(filename: Path) -> Mesh:
        return Mesh(
            str(filename),
            alpha=0.1,
            computeNormals=True,
            c=(1., 1., 1.)
        )

    def observe_atlas_mesh(self, change):
        print('saw atlas change')
        if self.model.atlas_mesh is None:
            self._atlas_mesh = Mesh()
        else:
            print('loading')
            worker = Task(self.load_mesh, self.model.atlas_mesh)
            worker.signals.finished.connect(partial(setattr, self, "atlas_mesh"))

            pool = QThreadPool.globalInstance()
            pool.start(worker)


    @warn_if_slow()
    def render(self, change):
        actors = [self._atlas_mesh]
        # box = self._atlas_mesh.box().wireframe().alpha(0.2).color((255, 0, 0))

        # actors.append(box)
        if len((points := self.model.points).coords) > 0:
            coords = points.coords
            colors = (np.hstack((points.colors, points.alphas)) * 255).astype(int)  # alphas needed for fast rendering.
            actors.append(Points(coords, r=3, c=colors))

        self.plotter.show(actors, at=0)
        self.plotter.addInset(self._atlas_mesh, pos=(.9, .9), size=0.1, c='w', draggable=True)
        # note: look at from vedo.applications import SlicerPlotter for inspiration


from pathlib import Path

import pandas as pd
from PyQt5.QtWidgets import QAction, QFileDialog

from regexport.model import AppState


class SaveCellsActionModel:
    text = "3. Save Cells"

    def register(self, model: AppState):
        self.model = model

    def savedata(self, filename):
        print('File saving...')

        types = {
            'Image': 'category',
            'BrainRegion': 'category',
            'Acronym': 'category',
            'X': 'float32',
            'Y': 'float32',
            'Z': 'float32',
        }

        types.update({col: 'uint16' for col in self.model.cells.columns if "Num Spots" in col})
        df = self.model.cells.astype(types)
        df: pd.DataFrame = df.drop(columns=['BGIdx'])

        print(df.info())
        print(filename)

        if filename.suffix.lower() == ".csv":
            df.to_csv(filename, index=False)
        elif filename.suffix.lower() == ".feather":
            df.to_feather(filename)
        else:
            raise TypeError(f"Error saving file {str(filename)}: {filename.suffix} extension not supported.")
        print("File saved")


class SaveCellsAction(QAction):

    def __init__(self, model: SaveCellsActionModel, *args, **kwargs):
        self.model = model
        super().__init__(*args, **kwargs)
        self.setText(model.text)
        self.triggered.connect(self.click)

    def click(self):
        filename, filetype_filter = QFileDialog.getSaveFileName(filter="Feather file (*.feather);;CSV file (*.csv)")

        if not filename:
            return

        self.model.savedata(Path(filename))



from typing import List

from PyQt5.QtWidgets import QComboBox
from traitlets import HasTraits, UseEnum, link

from regexport.model import AppState, AnalysisType
from regexport.views.utils import HasWidget


class AnalysisSelectorModel(HasTraits):
    analysis_type = UseEnum(AnalysisType, default_value=AnalysisType.RegionLabel)

    def register(self, model: AppState):
        link(
            (self, 'analysis_type'),
            (model, 'analysis_type')
        )


class AnalysisSelectorView(HasWidget):
    def __init__(self, model: AnalysisSelectorModel):
        self.dropdown = QComboBox()
        HasWidget.__init__(self, widget=self.dropdown)

        self.model = model
        self.set_dropdown_values(values=[analysis.name for analysis in self.model.analysis_type.__class__])
        self.model.observe(self.observe_analysis_type, ['analysis_type'])
        self.dropdown.currentTextChanged.connect(self.select_cmap_from_dropdown)

    def set_dropdown_values(self, values: List[str]):
        self.dropdown.clear()
        for value in values:
            self.dropdown.addItem(value)

    def observe_analysis_type(self, changed):
        self.dropdown.setCurrentText(self.model.analysis_type.name)

    def select_cmap_from_dropdown(self, text):
        print('selected', text)
        self.model.analysis_type = text


from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import Tuple, Optional

import numpy as np
import pandas as pd
from PyQt5.QtCore import QThreadPool
from bg_atlasapi import BrainGlobeAtlas
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap
from pandas import Series
from traitlets import HasTraits, Instance, directional_link

from regexport.model import AppState
from regexport.utils.parallel import Task
from regexport.utils.profiling import warn_if_slow


def is_parent(id: Series, selected_ids: Tuple[int], atlas: BrainGlobeAtlas) -> bool:
    assert len(id.unique()) == 1
    id = int(id.values[0])
    if id != 0:
        if id in selected_ids:
            return True
        else:
            is_ancestor = atlas.hierarchy.is_ancestor
            return any(is_ancestor(selected_id, id) for selected_id in selected_ids)
    else:
        return False


@dataclass(frozen=True)
class PointCloud:
    coords: np.ndarray = field(default=np.empty((0, 3), dtype=float))
    colors: np.ndarray = field(default=np.empty((0, 3), dtype=float))
    alphas: np.ndarray = field(default=np.empty((0, 1), dtype=float))

    def __post_init__(self):
        assert self.coords.ndim == 2 and self.coords.shape[1] == 3
        assert self.colors.ndim == 2 and self.colors.shape[1] == 3
        assert self.alphas.ndim == 2 and self.alphas.shape[1] == 1


class PlotterModel(HasTraits):
    atlas_mesh = Instance(Path, allow_none=True)
    points = Instance(PointCloud, default_value=PointCloud())

    def register(self, model: AppState):
        self.model = model
        directional_link(
            (model, 'atlas'),
            (self, 'atlas_mesh'),
            lambda atlas: Path(str(atlas.structures[997]['mesh_filename'])) if atlas is not None else None
        )
        model.observe(self.link_cells_to_points, names=[
            'selected_region_ids', 'cells', 'selected_colormap'
        ])

    def link_cells_to_points(self, change):
        model = self.model
        worker = Task(
            self.plot_cells,
            cells=model.cells,
            selected_region_ids=model.selected_region_ids,
            atlas=model.atlas,
            cmap=self.model.selected_colormap
        )
        worker.signals.finished.connect(partial(setattr, self, "points"))
        pool = QThreadPool.globalInstance()
        pool.start(worker)

    @staticmethod
    @warn_if_slow()
    def plot_cells(cells: Optional[pd.DataFrame], selected_region_ids: Tuple[int],
                   atlas: BrainGlobeAtlas, cmap: str = 'tab20c') -> PointCloud:
        if cells is None:
            return PointCloud()
        print('plotting')
        df = cells.copy(deep=False)
        print(df.head(), df.columns, sep='\n')
        cmap: ListedColormap = getattr(plt.cm, cmap)
        df[['red', 'green', 'blue', 'alpha']] = pd.DataFrame(cmap((codes := df.BrainRegion.cat.codes) / codes.max())[:, :4])
        if selected_ids := selected_region_ids:
            df['isSelected'] = (
               df
               .groupby('BGIdx', as_index=False)
               .BGIdx.transform(is_parent, selected_ids=selected_ids, atlas=atlas)
            )
            df = df[df['isSelected']]
            if len(df) == 0:
                return PointCloud()

        points = PointCloud(
            coords=df[['X', 'Y', 'Z']].values * 1000,
            colors=df[['red', 'green', 'blue']].values,
            alphas=df[['alpha']].values,
        )
        return points


import sys

from PyQt5.QtWidgets import QApplication

from regexport.actions.load_cells import LoadCellsAction, LoadCellsModel
from regexport.actions.load_atlas import LoadAtlasActionModel, LoadAtlasAction
from regexport.actions.save_cells import SaveCellsAction, SaveCellsActionModel
from regexport.actions.save_script import SaveGroovyScriptAction, SaveGroovyScriptActionModel
from regexport.views.analysis_selector import AnalysisSelectorModel, AnalysisSelectorView
from regexport.views.histogram import HistogramView, HistogramModel
from regexport.views.main_window import MainWindow
from regexport.model import AppState
from regexport.views.plot_3d import PlotterModel
from regexport.views.plot_3d.view import PlotterView
from regexport.views.region_tree import BrainRegionTree, BrainRegionTreeViewModel
from regexport.views.sidebar import Sidebar, SidebarModel
from regexport.views.text_selector import TextSelectorModel, DropdownTextSelectorView


def main():
    global model
    model = AppState()
    app = QApplication(sys.argv)
    plotter_model = PlotterModel()
    plotter_model.register(model=model)
    brain_region_tree_model = BrainRegionTreeViewModel()
    brain_region_tree_model.register(model=model)
    load_atlas_action_model = LoadAtlasActionModel()
    load_atlas_action_model.register(model=model)
    load_cells_action_model = LoadCellsModel()
    load_cells_action_model.register(model=model)
    save_cells_action_model = SaveCellsActionModel()
    save_cells_action_model.register(model=model)
    colormap_selector_model = TextSelectorModel()
    colormap_selector_model.register(
        model=model,
        options_attr='colormap_options',
        selected_attr='selected_colormap',
    )
    analysis_selector_model = AnalysisSelectorModel()
    analysis_selector_model.register(model=model)
    sidebar_model = SidebarModel()
    histogram_model = HistogramModel()
    histogram_model.register(model=model)
    win = MainWindow(
        main_widgets=[
            BrainRegionTree(model=brain_region_tree_model),
            PlotterView(model=plotter_model),
            Sidebar(
                model=sidebar_model,
                widgets=[
                    # AnalysisSelectorView(model=analysis_selector_model),
                    # DropdownTextSelectorView(model=colormap_selector_model),
                    # HistogramView(model=histogram_model),
                ]
            ),
        ],
        menu_actions=[
            SaveGroovyScriptAction(model=SaveGroovyScriptActionModel()),
            LoadAtlasAction(model=load_atlas_action_model),
            LoadCellsAction(model=load_cells_action_model),
            SaveCellsAction(model=save_cells_action_model),
        ]
    )
    win.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import Tuple, List

import numpy as np
import pandas as pd
from bg_atlasapi import BrainGlobeAtlas


def read_detection_file(filename: Path, atlas: BrainGlobeAtlas) -> pd.DataFrame:
    df = pd.read_csv(filename, sep='\t')
    print(df.columns)
    return (
        df
            .astype({
                'Allen CCFv3 X mm': float,
                'Allen CCFv3 Y mm': float,
                'Allen CCFv3 Z mm': float,
        })
            .pipe(lambda df: df.assign(
                BGIdx=get_brain_region_label(df['Allen CCFv3 X mm'].values, df['Allen CCFv3 Y mm'].values,
                                             df['Allen CCFv3 Z mm'].values, annotation=atlas.annotation,
                                            resolution=atlas.resolution)))
            .merge(atlas.lookup_df, right_on="id", left_on="BGIdx", how="left")
            .rename(columns={
                'Allen CCFv3 X mm': 'X',
                'Allen CCFv3 Y mm': 'Y',
                'Allen CCFv3 Z mm': 'Z',
                'name': 'BrainRegion',
                'acronym': 'Acronym',
        })
            .drop(columns=['id'])
            .assign(Image=filename.name)
    )  # .sample(10000)


def get_brain_region_label(x, y, z, annotation: np.ndarray, resolution: Tuple[int, int, int]) -> np.ndarray:
    sx, sy, sz = annotation.shape
    rx, ry, rz = resolution
    return annotation[
        np.clip((x / (rx / 1000)).astype(int), 0, sx - 1),
        np.clip((y / (ry / 1000)).astype(int), 0, sy - 1),
        np.clip((z / (rz / 1000)).astype(int), 0, sz - 1),
    ]


def read_detection_files(filenames: List[Path], atlas):
    df = pd.concat((read_detection_file(filename=filename, atlas=atlas) for filename in filenames), ignore_index=True)
    df = df.astype({'Acronym': 'category', 'BrainRegion': 'category', 'Image': 'category'})
    return df



from enum import Enum, auto

import pandas as pd
from bg_atlasapi import BrainGlobeAtlas
from matplotlib import pyplot as plt
from traitlets import HasTraits, Instance, observe, Tuple, List, UseEnum, Unicode


class AnalysisType(Enum):
    RegionLabel = auto()
    SubCellCount = auto()


class AppState(HasTraits):
    atlas = Instance(BrainGlobeAtlas, allow_none=True)
    cells = Instance(pd.DataFrame, allow_none=True)
    selected_region_ids = Tuple(default_value=())  # should be tuple of ints
    colormap_options = List(Unicode(), default_value=[cmap for cmap in plt.colormaps() if not cmap.endswith('_r')])#['tab20c', 'viridis'])
    selected_colormap = Unicode(default_value='tab20c')
    analysis_type = UseEnum(AnalysisType, default_value=AnalysisType.RegionLabel)

    @observe('selected_region_ids')
    def _on_change_selected_region_ids(self, change):
        print(f"Selected: {change['new']}")

    @observe('selected_colormap')
    def _on_colormap_change(self, change):
        print('changed colormap to', self.selected_colormap)